const ClassMethod = require("../types/ClassMethod");
const { types: t, template, traverse } = require("@babel/core");
const { findByPattern, ucfrist, toCamelCase } = require("../utils");
const _ = require("underscore");
const ClassProperty = require("../types/ClassProperty");
const CallExpression = require("../types/CallExpression");
const { ObjectProperty } = require("../types");
class BaseScript {
    constructor(ast) {
        this.ast = ast;
        // this.transformFunctionParams();
        this.transformRequest();
        this.transformUtils();
        this.transformAssert();
        this.transformComment();
        this.transformAlias();
    }

    transformAlias() {
        let alias = {};
        traverse(this.ast, {
            ObjectProperty(path) {
                let objectProperty = new ObjectProperty(path);
                let comments = objectProperty.getLeadingComments();
                if (_.isArray(comments) && !_.isEmpty(comments)) {
                    comments.forEach(function (item) {
                        let match = item.value.match(/@alias ([A-Za-z]+)/);
                        if (_.isArray(match) && match[1]) {
                            item.value.replace("alias", "name");
                            alias[objectProperty.getName().name] = match[1];
                        }
                    });
                }
            },
        });
        traverse(this.ast, {
            Identifier(path) {
                if (_.keys(alias).includes(path.node.name)) {
                    path.node.name = alias[path.node.name];
                }
            },
        });
    }

    transformComment() {
        traverse(this.ast, {
            ClassProperty(path) {
                let classProperty = new ClassProperty(path);
                classProperty.getNode().trailingComments = null;
            },
            ClassMethod(path) {
                let classMethod = new ClassMethod(path);
                classMethod.getNode().trailingComments = null;
            },
        });
    }

    transformAssert() {
        traverse(this.ast, {
            ClassMethod(path) {
                let classMethod = new ClassMethod(path);
                // let comment = classMethod.getLeadingComments();
                let params = classMethod.getParams();
                if (_.isEmpty(params)) {
                    return;
                }
                params.forEach(function (item) {
                    // let pattern = new RegExp(`\@param {(.*?)} ${item.name}`);
                    // let match = comment.value.match(pattern);
                    if (t.isTypeAnnotation(item.typeAnnotation)) {
                        let type;
                        if (
                            t.isStringTypeAnnotation(
                                item.typeAnnotation.typeAnnotation
                            )
                        ) {
                            type = "string";
                        } else if (
                            t.isNumberTypeAnnotation(
                                item.typeAnnotation.typeAnnotation
                            )
                        ) {
                            type = "number";
                        } else if (
                            t.isObjectTypeAnnotation(
                                item.typeAnnotation.typeAnnotation
                            )
                        ) {
                            type = "object";
                        } else if (
                            t.isArrayTypeAnnotation(
                                item.typeAnnotation.typeAnnotation
                            )
                        ) {
                            type = "array";
                        }
                        classMethod.prepend([
                            template(
                                `
                                if (!\_.is${ucfrist(type)}(${item.name})) {
                                    throw new Error("The type of the ${
                                        item.name
                                    } parameter of the ${classMethod.getName()} method must be ${type}.");
                                }
                            `,
                                {
                                    placeholderPattern: /^[$A-Z0-9]+$/,
                                }
                            )(),
                        ]);
                    }
                });
            },
        });
    }

    /**
     * 转换工具函数
     */
    transformUtils() {
        traverse(this.ast, {
            CallExpression(path) {
                let callExpression = new CallExpression(path);
                ["Array", "Object", "String"].forEach(function (value) {
                    if (callExpression.getName() === `isEmpty${value}`) {
                        callExpression.getPath().replaceWith(
                            template(
                                `_.is${value}(%%target%%) && _.isEmpty(%%target%%)`
                            )({
                                target: callExpression.getArguments(),
                            })
                        );
                    }
                });
                if (callExpression.getName()) {
                    if (callExpression.getName().indexOf("findBy") !== -1) {
                        callExpression.getPath().replaceWith(
                            template(
                                `%%object%%.find(item => item.%%field%% === %%value%%)`
                            )({
                                object: callExpression.getObject(),
                                field: t.Identifier(
                                    toCamelCase(
                                        callExpression
                                            .getName()
                                            .replace("findBy", "")
                                    )
                                ),
                                value: callExpression.getArguments()[0],
                            })
                        );
                    }
                    if (
                        callExpression.getName().indexOf("findIndexBy") !== -1
                    ) {
                        callExpression.getPath().replaceWith(
                            template(
                                `%%object%%.findIndex(item => item.%%field%% === %%value%%)`
                            )({
                                object: callExpression.getObject(),
                                field: t.Identifier(
                                    toCamelCase(
                                        callExpression
                                            .getName()
                                            .replace("findIndexBy", "")
                                    )
                                ),
                                value: callExpression.getArguments()[0],
                            })
                        );
                    }
                    if (callExpression.getName().indexOf("deleteBy") !== -1) {
                        callExpression.getPath().replaceWith(
                            template(
                                `%%object%%.splice(%%object%%.findIndex(item => item.%%field%% === %%value%%), 1)`
                            )({
                                object: callExpression.getObject(),
                                field: t.Identifier(
                                    toCamelCase(
                                        callExpression
                                            .getName()
                                            .replace("deleteBy", "")
                                    )
                                ),
                                value: callExpression.getArguments()[0],
                            })
                        );
                    }
                }
            },
            Identifier(path) {
                if (
                    [
                        "isEmpty",
                        "isString",
                        "isArray",
                        "isNumber",
                        "isBoolean",
                    ].includes(path.node.name)
                ) {
                    let parent = path.findParent((path) =>
                        path.isCallExpression()
                    );
                    if (parent && t.isIdentifier(parent.node.callee)) {
                        parent.replaceWith(
                            template(`_.${path.node.name}(%%value%%)`)({
                                value: parent.node.arguments[0],
                            })
                        );
                    }
                }
            },
        });
    }

    // transformFunctionParams() {
    //     traverse(this.ast, {
    //         ClassMethod(path) {
    //             let classMethod = new ClassMethod(path);
    //             let values = [];
    //             if (classMethod.getParams().length === 0) {
    //                 return;
    //             }
    //             classMethod.getParams().forEach(function (item) {
    //                 item.typeAnnotation = null;
    //                 values.push(item.name);
    //             });

    //             classMethod.prepend([
    //                 template(`
    //                     [${values.join(",")}].forEach(function(value, key){
    //                         if([undefined, null, ""].includes(value)) {
    //                             throw new Error(\`\${["${values.join(
    //                                 '","'
    //                             )}"][key]} parameter of ${
    //                     classMethod.getName().name
    //                 } method is required.\`)
    //                         }
    //                     })
    //                 `)(),
    //             ]);
    //         },
    //     });
    // }

    /**
     * 转换请求方法
     * @example
     * // 编译前：
     * request.post("/path/to/backend", {foo: "bar"}).then(function() {
     *      // doSomething...
     * })
     * // 编译后：
     * $.ajax({
     *      url: "/path/to/backend",
     *      method: "POST",
     *      contentType: "application/json",
     *      dataType: "json",
     *      data: JSON.stringify({foo: "bar"}),
     *      success: function() {
     *          // doSomething...
     *      }
     * })
     */
    transformRequest() {
        traverse(this.ast, {
            ExpressionStatement(path) {
                if (findByPattern("request.post", path.node)) {
                    let params = path.node.expression.callee.object.arguments;
                    let func = path.node.expression.arguments[0];
                    let data = "";
                    let placeholder = {
                        url: params[0],
                        success: func,
                    };
                    if (params[1] && params[1].properties.length > 0) {
                        data = "data: JSON.stringify(%%data%%),";
                        placeholder["data"] = params[1] || {};
                    }
                    path.replaceWith(
                        template(
                            `$.ajax({
                            url: %%url%% ,
                            method: "POST",
                            contentType: "application/json",
                            dataType: "json",
                            ${data}
                            success: %%success%%
                        })`
                        )(placeholder)
                    );
                }
            },
        });
    }
}

module.exports = BaseScript;
